/* DCPRAG.C */
/* Y.L. Noyelle, Supelec, France 1994 */
/* Management of d-pragmas */

#include <stdlib.h>
#include <setjmp.h>
#include <string.h>
#include "dcprag.h"
#include "dcprag.ph"
#include "dcblk.h"
#include "dcdecl.h"
#include "dcdir.h"
#include "dcdir.ph"
#include "dcexp.h"
#include "dcext.h"
#include "dcmain.h"
#include "dcrec.h"
#include "dcrec.ph"
#include "dcrecdir.ph"

#ifdef VMS
#  pragma noinline (procPriv)
#endif

struct _TCBlk {
  const TsemanElt *lTypeId, *rTypeId;
  TpcTypeElt resulType;
  TlineNb defLineNb;
  Tstring defFileName;
  TCBlk *next;
};

struct _TmngtPriv {
  TmngtPriv *prev, *chain;
  const Tchar fileName[1];  /* stub for name, so must be at end */
};
/*~zif Offset(TmngtPriv, fileName) != sizeof(TmngtPriv)-1
				      "Field 'fileName' not at end of struct" */
DefRepresType

/* Function profiles */
static TCBlk *allocTCBlk(void), *freeTCBlk(TCBlk *);
/*static Tstring dpName1(Tdprag, bool);*/
static bool getTypeIdent(TpcTypeElt *, bool);
static TtypeSort intPromotion(TpcTypeElt);
static void addTCB(TkOpTC, TpcTypeElt, TpcTypeElt, TpcTypeElt),
  initWarnMngt(void),
  manageTC(TkOpTC, TpcTypeElt, TpcTypeElt, TpcTypeElt, bool),
  procPriv(TlitString), skipToComma(void),
  verifyTC(TpcTypeElt, TpcTypeElt, TpcTypeElt);

/* Global variables */
static TmngtPriv *headPrivBlks = NULL;

/* External variables */
uint ctrTCB = 0;  /* count of TypeCombination blocks */
TmngtPriv *headListPrivTo;
bool heedDP = False;
bool noWarn = False;
TCBlk *(*pHeadTCBLists)[ModAsgn - AddAsgn + 1 /*~IndexType TkOpTC*/] = NULL;

static Tname tabDPragN[] = {
	/* Non-token, or token with possible argument => '\1' at end (cf also
	   dpName() ). */
	/* Sorted by order of increasing h-codes (see tabDPragHC) */
	ConvTname("\x3\0zif\1"),
	ConvTname("\x4\0Init\1"),
	ConvTname("\x4\0Warn\1"),
	ConvTname("\x5\0Undef\1"),
	ConvTname("\x6\0CastTo\1"),
	ConvTname("\x6\0Ignore"),
	ConvTname("\x6\0NoWarn\1"),
	ConvTname("\x6\0Public\1"),
	ConvTname("\x6\0SizeOK"),
	ConvTname("\x7\0Generic\1"),
	ConvTname("\x7\0DynInit"),
	ConvTname("\x7\0Masking"),
	ConvTname("\x7\0OddCast"),
	ConvTname("\x7\0NoBreak"),
	ConvTname("\x7\0NotUsed"),
	ConvTname("\x7\0PopWarn\1"),
	ConvTname("\x7\0Private\1"),
	ConvTname("\x7\0Utility"),
	ConvTname("\x8\0ExactCmp"),
	ConvTname("\x8\0FullEnum"),
	ConvTname("\x8\0LocalAdr"),
	ConvTname("\x8\0RootType\1"),
	ConvTname("\x8\0UndefTag\1"),
	ConvTname("\x9\0IndexType\1"),
	ConvTname("\x9\0MayModify"),
	ConvTname("\x9\0NoDefault\1"),
	ConvTname("\x9\0SameValue"),
	ConvTname("\x9\0ResultPtr"),
	ConvTname("\x9\0PrivateTo\1"),
	ConvTname("\xA\0BackBranch"),
	ConvTname("\xA\0DollarSign"),
#ifdef TstAdjFiles
	ConvTname("\xA\0EndAdjFile\1"),
	ConvTname("\xA\0EndSysFile\1"),
#endif
	ConvTname("\xA\0LiteralCst"),
	ConvTname("\xA\0PortableQM"),
	ConvTname("\xA\0ResultType"),
	ConvTname("\xA\0PseudoVoid\1"),
#ifdef TstAdjFiles
	ConvTname("\xA\0SimAdjFile\1"),
	ConvTname("\xA\0SimSysFile\1"),
#endif
	ConvTname("\xB\0NonConstExp"),
	ConvTname("\xB\0VoidToOther"),
	ConvTname("\xC\0DccCompliant\1"),
	ConvTname("\xC\0ComposingHdr"),
	ConvTname("\xC\0NeverReturns\1"),
	ConvTname("\xC\0SideEffectOK"),
	ConvTname("\xC\0SizeOfMemBlk"),
	ConvTname("\xF\0TypeCombination\1")
	/*~ zif __index != __extent(Tdprag) "Bad 'tabDPragN'" */};

static ThCode tabDPragHC[] = {
        0x6FA,			/* zif */
        0xEA6,			/* Init */
        0x145E,			/* Warn */
        0x65EB,			/* Undef */
        0x10AE0,		/* CastTo */
        0x16044,		/* Ignore */
        0x1A733,		/* NoWarn */
        0x1D18F,		/* Public */
        0x1E375,		/* SizeOK */
        0x64A30,		/* Generic */
        0x67C28,		/* DynInit */
        0x795F1,		/* Masking */
        0x7FEC7,		/* OddCast */
        0x817AF,		/* NoBreak */
        0x8794B,		/* NotUsed */
        0x8E91C,		/* PopWarn */
        0x908DA,		/* Private */
        0xA4E6B,		/* Utility */
        0x20F2ED,		/* ExactCmp */
        0x21F8FC,		/* FullEnum */
        0x27351A,		/* LocalAdr */
        0x2F194E,		/* RootType */
        0x31C772,		/* UndefTag */
        0xB18D4D,		/* IndexType */
        0xBDF8C6,		/* MayModify */
        0xCA3005,		/* NoDefault */
        0xDF8C02,		/* SameValue */
        0xE0BA94,		/* ResultPtr */
        0xE1DE0D,		/* PrivateTo */
        0x257CA39,		/* BackBranch */
        0x2F2FBD1,		/* DollarSign */
#ifdef TstAdjFiles
        0x2F8826E,		/* EndAdjFile */
        0x2FDE362,		/* EndSysFile */
#endif
        0x3C62E49,		/* LiteralCst */
        0x4618D4B,		/* PortableQM */
        0x463A773,		/* ResultType */
        0x46A6080,		/* PseudoVoid */
#ifdef TstAdjFiles
        0x486AB24,		/* SimAdjFile */
        0x48C0C18,		/* SimSysFile */
#endif
        0x1476A396,		/* NonConstExp */
        0x191BE16A,		/* VoidToOther */
        0x40F23DAD,		/* DccCompliant */
        0x4718E851,		/* ComposingHdr */
        0x621221EB,		/* NeverReturns */
        0x70AFBE08,		/* SideEffectOK */
        0x734A69AC,		/* SizeOfMemBlk */
        0xAC6C71FA		/* TypeCombination */
	/*~ zif __index != __extent(Tdprag) "Bad 'tabDPragHC'" */};

static bool lastIsPopWarnAtMainLvl = True;  /* True because of possible
							       option '-zwa'. */
static uint warnStk = 0;

static void initWarnMngt(void)
{
  lastIsPopWarnAtMainLvl = True;
  warnStk = 0;
}

TvalTok manageDPrag(void)
/* Done that way because of possible macro-expansion inside d-pragma */
{
  TvalTok resulTok;

  dpragNst++;
  if (! (macroExpand && (concatFl==CNoBit || concatFl & CShield) || heedDP)
							 ) resulTok.tok = DPRAG;
  else {  /* search d-pragma */
    register int i = 0, j = NbElt(tabDPragHC) - 1, k;
    TtokLvl savGetTokLvl = getTokLvl;

    if (!dirLineFl && (getTokLvl = tokLvl0 - 1)==0) storeTokTxt(DPragPrefix);
    resulTok.tok = WHITESPACE;
    if (NxtTok() != IDENT) {
      if (curTok.tok != ENDDPRAG) err0(DPragNameExptd);
      goto exitL1;}
    do {
      k = (i + j) >> 1;
      if (curTok.hCod > tabDPragHC[k]) {i = k + 1; continue;}
      if (curTok.hCod < tabDPragHC[k]) {j = k - 1; continue;}
      if (memcmp(tabDPragN[k], curTok.IdName, FullLgt(curTok.IdName)) != 0) {
        j = k - 1;  /* arbitrary */
        continue;}
      if (*(tabDPragN[k]+FullLgt(tabDPragN[k])) == 0) {  /* token-like d-p. */
        resulTok.tok = DPTOK;
        resulTok.val = (uint)k;}
      else {  /* d-pragma to be interpreted */
        Tdprag curDP = (Tdprag)k;

        switch (curDP) {
          case CASTTO: case INDEXTYPE:
            GetNxtTok();
            resulTok.dpType = declType();
            goto rootTypL;
          case DCCCOMPL:
            if (!sysHdrFile) err0(IlgDccCmpl | Warn1);
            else sysHdrFile = sysAdjHdrFile = False;
            break;
#ifdef TstAdjFiles
          case ENDADJ: adjustFile = False; sysAdjHdrFile = sysHdrFile; break;
          case ENDSYS: sysHdrFile = False; sysAdjHdrFile = adjustFile; break;
#endif
       /* case GENERIC: see PSEUDOVOID */
          case INIT:
            do {
              TsemanElt *ptrId;

              GetNxtTok();
              if (curTok.tok != IDENT) err0(IdExptd);
              else {
                if ((ptrId = curTok.ptrSem) == NULL) errWN(UndefId, curTok.
									IdName);
                else if (ptrId->kind!=Obj || ptrId->attribb<Auto && ptrId->
			       attribb!=StatiL)  errWN(NotVarId, curTok.IdName);
                else if (ptrId->initBefUsd || ptrId->initlz || ptrId->used
				    ) errWN(UslInitBU|Warn1|Rdbl, ptrId->nameb);
                else if (adjMacro) ptrId->initlz = True;
                else ptrId->initBefUsd = True;
                GetNxtTok();}
            } while (curTok.tok == COMMA);
            goto exitL;
       /* case INDEXTYPE: see CASTTO */
       /* case NEVRET: see PSEUDOVOID */
          case NODEFAULT: resulTok.tok = DEFAULT; resulTok.hCod = 0; break;
       /* case NOWARN:  see WARN */
          case POPWARN:
            noWarn = warnStk & 1;
            warnStk >>= 1;
            if (dpHereOrAloneInMac() && !insideIncldFile()
						) lastIsPopWarnAtMainLvl = True;
            break;
          case GENERIC:
          case NEVRET:
          case PSEUDOVOID:
            if (NxtTok() != IDENT) goto rootTypL;
            do {
              TpcTypeElt type;
              const TsemanElt *ptrId = curTok.ptrSem;

              if (! adjustFile) goto ilgDPL;
              if (curTok.tok != IDENT) {err0(IdExptd); continue;}
              else if (ptrId != NULL) {
                if (ptrId->kind!=Obj || (type = ptrId->type)==NULL || !IsFct(
					  type)) errWN(NotFctId, curTok.IdName);
                else {
                  if (curDP == GENERIC) {
                    const TdeclElt *parElt;

                    for (parElt = type->paramList; parElt != NULL; parElt =
    								    parElt->cdr)
                      if (parElt->parQal == ResulTypV) goto ilgDPL;
                    ((ModifType)type)->generic = True;
                    makeGeneric((ModifType)type);}
                  else {
                    TpcTypeElt retType = NxtTypElt(type);

                    if (retType==NULL || (retType->typeSort == Void)!=
						 (curDP == NEVRET)) goto ilgDPL;
                    ((ModifType)type)->pvNr = True;}}}
              goto skip1L;
ilgDPL:
              errIlgDP(curDP);
skip1L:
              GetNxtTok();
            } while (Found(COMMA));
            goto exitL;
          case PRIVATE:
          case PRIVATETO:
          case PUBLIC:
            if (! headerFile) errWS(DPragNotAlwd | Warn2, dpName(curDP));
            headListPrivTo = NULL;  /* do not free preceding list, which is
			      probably refered to by 'listAlwdFiles' members. */
            if (curDP == PUBLIC) break;
            if (curDP == PRIVATE) {
              if (headerFile) {
                const Tchar *posDot = NULL;

                enterBlock();  /* for string storage reclaiming purpose */
                for (;;) {
                  register const Tchar *ptrC = NakedFName(curFileName) - 1;
                  TlitString fileName = ptrFreeIdSpace();

                  do {storeStrCh(*++ptrC);} while (ptrC!=posDot && *ptrC!='\0');
                  if (posDot != NULL) {  /* creates corresponding body file
									name. */
                    ptrC = &CSuffix[0];
                    do {storeStrCh(*++ptrC);} while (*ptrC != '\0');}
                  procPriv(fileName);
                  if (posDot != NULL) break;
                  posDot = SearchDot(NakedFName(curFileName));}
                exitBlock();}  /* because of 'getLitString()' */
              break;}
            /* else PRIVATETO */
            GetNxtTok();
            do {
              TlitString str;

              if ((str = getLitString())!=NULL && headerFile) procPriv(str);
              exitBlock();  /* because of 'getLitString()' */
            } while (Found(COMMA));
            goto exitL;
          case ROOTTYP:
            if (NxtTok() == IDENT) {
              do {
                TpcTypeElt type;

                if (getTypeIdent(&type, False) && type!=NULL)
                  if (!type->NxtIsTypId || !adjustFile) errIlgDP(ROOTTYP);
                  else ((ModifType)type)->rootTyp = True;
              } while Found(COMMA);
              goto exitL;}
rootTypL:
/*            if (adjustFile) errIlgDP(tok); else */
            resulTok.tok = DPTOK;
            resulTok.val = (uint)curDP;
            goto exitL;
#ifdef TstAdjFiles
          case SIMADJ: sysAdjHdrFile = adjustFile = True; break;
          case SIMSYS: sysAdjHdrFile = sysHdrFile = True; break;
#endif
          case TYPECMBN:
            if (nestLvl != 0) err0(TCNotAtLvl0);
            do {  
              bool commut = True;
              TkOpTC oper;
              TpcTypeElt lType, rType, resType;
	    
              if (getTypeIdent(&lType, True)) {
                if (curTok.tok==TILDE) {
                  commut = False;
                  GetNxtTok();}
                oper = (curTok.tok == ADDOP)
                       ? AddTC + ((TkAdd)curTok.val - Add)
                       : (curTok.tok == STAR)
                         ? MulTC
                         : (curTok.tok == MULOP)
                           ? MulTC + ((TkMul)curTok.val - Mul)
                           : (err0(IlgTCOptr), IlgTcOp);
                if (oper >= AddTC) {
                  if (! (oper==AddTC || oper==MulTC)) 
                    if  (! commut) err0(NonCommutOper);
                    else commut = False;
                  if (getTypeIdent(&rType, True)) {
                    if (curTok.tok != ARROW) err0(ArrowExptd);
                    else {
                      if (oper==DivTC && (CompatType(lType, rType, NoCheck) ||
		  	  CompatType(rType, lType, NoCheck))) {  /* one is
						      ascendant of the other. */
                        GetNxtTok();
                        if ((resType = declType())!=NULL && (resType->noOwner ||
					       !IsTypeSort(resType, NumEnum))) {
                          errWT(IlgType, resType);
                          FreeExpType2(resType);
                          resType = NULL;}}
                      else if (! getTypeIdent(&resType, True)) goto skipL;
                      manageTC(oper, lType, rType, resType, commut);
                      if (curTok.tok!=COMMA && curTok.tok!=ENDDPRAG) errWS(
						   CommaOr, errTxt[EndDPExptd]);
skipL:;             }}}}
              skipToComma();
            } while (curTok.tok == COMMA);
            goto exitL; 
          case UNDEF: case UNDEFTAG:
            if (nestLvl != 0) err0(IllUndef);
            do {
              TsemanElt *pId;

              if (NxtTok() != IDENT) {
                err0(IdExptd);
                skipToComma();}
              else {
                if (nestLvl == 0) {
                  if (curDP == UNDEFTAG) {
                    *(TnameNC /*~OddCast*/)(curTok.IdName + DispNSId) =
							    (TnameAtom)TagSpace;
                    pId = searchSymTab(curTok.IdName);}
                  else
                    pId = curTok.ptrSem;
                  if (pId == NULL) errWN(UnknId, curTok.IdName);
                  else if (pId->undef && curDP==UNDEFTAG) errWN(AlrdUndef |
							  Warn2, curTok.IdName);
                  else pId->undef = True;}
                GetNxtTok();}
            } while (curTok.tok == COMMA);
            goto exitL;
          case WARN: case NOWARN:
            warnStk = warnStk<<1 | (uint)noWarn;
            noWarn = (curDP == NOWARN);
            if (dpHereOrAloneInMac() && !insideIncldFile()) {
              if (!lastIsPopWarnAtMainLvl && !((warnStk & 1) ^ noWarn)) err0(
				(noWarn)? WarnAlrdOff|Warn2 : WarnAlrdOn|Warn2);
              lastIsPopWarnAtMainLvl = False;}
            break;
          case ZIF: {
              TresulExp saveCExp = cExp, saveLExp = lExp;
              TlitString str;
              bool trueZif, saveZifExp = zifExp, saveIfDirExp = ifDirExp;
              Trchbl savNSR = nxtStmtRchbl;

              zifExp = True;
              ifDirExp = False;
              GetNxtTok();
              trueZif = correctExprN(NULL, Bool, True, "zif", False) && !cExp.
							 errEvl && cExp.uVal!=0;
              if ((str = getLitString())!=NULL && trueZif) {
                Tchar *ptrZifMsg, *w;

                w = ptrZifMsg = allocPermSto(initGetStrLit(str));
                while ((*w++ = nxtStrLitChar()) != '\0') {}
                errWS(ZifWarn | Warn3, ptrZifMsg);
                freeLastPermSto();
                resetGetStrLit();}
              exitBlock();  /* because of 'getLitString()' */
              zifExp = saveZifExp;
              ifDirExp = saveIfDirExp;
              cExp = saveCExp; lExp = saveLExp; nxtStmtRchbl = savNSR;}
            goto exitL;
          default: sysErr(ExCod7);}}
      GetNxtTok();
exitL:
      if (curTok.tok != ENDDPRAG) err0(EndDPExptd);
      goto exitL1;
    } while (i <= j);  /* end binary search loop */
    err0(UnknDPrag);
exitL1:
    {
      static const Ttok zEndDP[] = {NoSwallowTok, ENDDPRAG, ENDDIR, EndSTok};

      skipTok(zEndDP);
    }
    if (curTok.tok == ENDDIR) {dpragNst--; dirLineFl = True; backUp();}  /* to
					       get ENDDIR back as next token. */
    getTokLvl = savGetTokLvl;}
  return resulTok;
}

/*~Undef lastIsPopWarnAtMainLvl,warnStk */

/******************************************************************************/
/*                                UTILITIES                                   */
/******************************************************************************/

static void addTCB(TkOpTC op, TpcTypeElt lType, TpcTypeElt rType, TpcTypeElt
									resType)
{
  TCBlk *pcTCB, *prec = NULL;
  register const TsemanElt *lTypId = lType->typeId;

  if (pHeadTCBLists == NULL) {
    /* Allocate and initialize table of TypeCombinations for each operator
       (Add -> Mod). */
    TCBlk **ptr;

    MyAlloc(pHeadTCBLists, sizeof(*pHeadTCBLists));
    for (ptr = &(*pHeadTCBLists)[AddTC]; ptr <= &(*pHeadTCBLists)[ModTC];)
      *ptr++ = NULL;}
  for (pcTCB = (*pHeadTCBLists)[op]; pcTCB != NULL; pcTCB = pcTCB->next) {
    if (pcTCB->lTypeId==lTypId && pcTCB->rTypeId==rType->typeId) {
      static Tchar tCOperRepre[] = " ";
#include "dctxttok.h"

      tCOperRepre[0] = txtAsgnTok[AddAsgn + (int)op][0];
      errWFName(TCAlrdDef, pcTCB->defLineNb, pcTCB->defFileName, lTypId->
			     nameb, nameToS(rType->typeId->nameb), tCOperRepre);
      longjmp(*curErrRet, 1);}
    prec = pcTCB;}
  {
    TCBlk *w = allocTCBlk();

    w->next = NULL;
    w->lTypeId = lTypId;
    w->rTypeId = rType->typeId;
    w->resulType = resType;
    w->defLineNb = lineNb;
    w->defFileName = curFileName;
    if (prec == NULL) (*pHeadTCBLists)[op] = w;
    else prec->next = w;}
}

static AllocXElt(allocTCBlk, TCBlk, ctrTCB, ;)

Tstring dpName(Tdprag pragNo)
{
  static Tstring listStrings[] = {DPragPrefix, NULL, DPragSuffix};
  const Tstring *w, *limW = /*&listStrings[NbElt(listStrings)
				      - 1] : */&listStrings[NbElt(listStrings)];
  register Tchar *ptrBuf = &tokTxtBuf[0];

  listStrings[1] = (Tstring)(tabDPragN[(int)pragNo] + LgtHdrId);
  for (w = &listStrings[0]; w != limW; ) {
    register Tstring w1 = *w++;

    while (*w1 > '\1') *ptrBuf++ = *w1++;}
  *ptrBuf = '\0';
  return &tokTxtBuf[0];
}

void freeTCB(void)
{
  if (pHeadTCBLists != NULL) {
    TCBlk **ptrInArr;

    for (ptrInArr = &(*pHeadTCBLists)[AddTC]; ptrInArr <= &(*pHeadTCBLists)[
								     ModTC]; ) {
      TCBlk *curTCB = *ptrInArr++;

      while (curTCB != NULL) curTCB = freeTCBlk(curTCB);}
    free(pHeadTCBLists);
    pHeadTCBLists = NULL;}
}

static FreeXElt(freeTCBlk, TCBlk *, ctrTCB, ; , next)

static bool getTypeIdent(TpcTypeElt *gottenType, bool fl)
{
  bool resul = False;
  const TsemanElt *ptrId;

  if (fl) GetNxtTok();
  if (curTok.tok == IDENT) {
    if ((ptrId = curTok.ptrSem) == NULL) {
      DefSem1(artifDef, Type, True, True, False);

      errWN(UndefId, curTok.IdName);
      (void)defineId(artifDef);
      *gottenType = NULL;
      resul = True;}
    else if (ptrId->kind == Type) {
      *gottenType = (fl && ptrId->namedType!=NULL && (!IsTypeSort(ptrId->
		namedType,NumEnum) || !ptrId->namedType->paralTyp))? (errWT(
		      TypeNotParal, ptrId->namedType), NULL) : ptrId->namedType;
      resul = True;}
    else errWN(NotTypeId, curTok.IdName);
    GetNxtTok();}
  else err0(IdExptd);
  return resul;
}

void initPrag(void)
{
  initWarnMngt();
  /* Free memory */
  while (headPrivBlks != NULL) {
    TmngtPriv *w = headPrivBlks;

    headPrivBlks = headPrivBlks->chain;
    free(w);}
  headListPrivTo = NULL;
}

static TtypeSort intPromotion(TpcTypeElt x)
{
  TtypeSort ts = x->typeSort;

  if (ts == Enum) return Int;
  if (ts < Int)
#if USHRT_MAXC == UINT_MAXC
    if (chkPortbl && ts==UShort) errWS(NonPortTC | Warn2, IntStr);
#endif
    return Int;
  return ts;
}

bool isFNameVisible(Tstring fileName, const TsemanElt *pTag)
{
  register const TmngtPriv *w = pTag->listAlwdFiles;

  fileName = skipPath(fileName);
  do {
    if (strJokerEq(w->fileName, fileName)) return True;
  } while ((w = w->prev) != NULL);
  return False;
}

static void manageTC(TkOpTC oper, TpcTypeElt lType, TpcTypeElt rType,
						TpcTypeElt resType, bool commut)
{
  jmp_buf localJmpBuf, *savErrRet;

  if (nestLvl!=0 || lType==NULL || rType==NULL) return; /* ignore combination */
  if (resType != NULL) {
    verifyTC(lType, rType, resType);
    if (oper == ModTC) {
      if (IsTypeSort(resType, Flt)) errWSTT(IlgTCModOpnd, NULL, lType, rType);
      if (lType != resType) err0(IncorTCModRes | Warn2 | PossErr);}}
  savErrRet = curErrRet;
  curErrRet = &localJmpBuf /*~LocalAdr*/;
  if (setjmp(*curErrRet) == 0) {
    addTCB(oper, lType, rType, resType);
    if (commut && rType!=lType) addTCB(oper, rType, lType, resType);
    if (resType != NULL)
      switch (oper) {
        case AddTC:
          if (resType == lType) addTCB(SubTC, lType, rType, resType);
							    /* 'generic' type */
          break;
        case MulTC:
          if (commut) {
            addTCB(DivTC, resType, lType, rType);
            if (rType != lType) addTCB(DivTC, resType, rType, lType);
            if (IsTypeSort(resType, WhoEnum)) {
              addTCB(ModTC, resType, lType, resType);
              if (rType != lType) addTCB(ModTC, resType, rType, resType);}}
          break;
        case DivTC:
          if (IsTypeSort(resType, WhoEnum)) addTCB(ModTC, lType, rType, lType);
          break;
        /*~NoDefault*/}}
  curErrRet = savErrRet;
}

static void procPriv(TlitString str)
{
  register Tchar *pc;
  Tstring sp;
  TmngtPriv *w;

  MyAlloc(w, initGetStrLit(str) + Offset(TmngtPriv, fileName));
  pc = (Tchar * /*~OddCast*/)&w->fileName[0];
  while ((*pc++ = nxtStrLitChar()) != '\0') {};  /* note indicated file name */
  resetGetStrLit();
  /* Suppress possible path part in file name */
  if ((sp = skipPath(w->fileName)) != &w->fileName[0]) {
    size_t lgt = strlen(sp) + 1;

    memmove((Tchar * /*~OddCast*/)&w->fileName[0], sp, lgt);
    w = realloc(w, lgt + Offset(TmngtPriv, fileName));}
  w->prev = headListPrivTo;
  headListPrivTo = w;
  w->chain = headPrivBlks;  /* for memory		    */
  headPrivBlks = w;	    /*		  freeing purposes. */
}

TpcTypeElt searchTC(TkOpTC oper)
/* Returns usually QuasiNULLval(TpcTypeElt) if no combination found */
{
  TpcTypeElt lType, rType;
  register const TsemanElt *lTypeId, *rTypeId;
  const TCBlk *ptrTCoper = (*pHeadTCBLists)[oper];

  if (ptrTCoper==NULL || (lType = lExp.type)==NULL || !lType->paralTyp ||
       cExp.type==NULL || !cExp.type->paralTyp) return QuasiNULLval(TpcTypeElt);
  /* Prepare to row up hierarchy, in order to find possible combination between
     ancestors. */
  do {
    lTypeId = lType->typeId;
    rType = cExp.type;
    do {
      register const TCBlk *pcTCB = ptrTCoper;

      rTypeId = rType->typeId;
      do {
        if (lTypeId==pcTCB->lTypeId && rTypeId==pcTCB->rTypeId) return
							       pcTCB->resulType;
      } while ((pcTCB = pcTCB->next) != NULL);
    } while (!rType->rootTyp && (rType = rTypeId->type)!=NULL && rType->
								      paralTyp);
  } while (!lType->rootTyp && (lType = lTypeId->type)!=NULL && lType->paralTyp);
  if (InsideInterval(oper, MulTC, DivTC) &&
  /* Search whether one operand appears in a type combination involving opera-
     tor '*' ('normal' type combination for non dimensionless type). If so,
     complain about missing typeCombination (cf Volt*Volt/Ohm). */
		     ((ptrTCoper = (*pHeadTCBLists)[MulTC])!=NULL || (ptrTCoper
					    = (*pHeadTCBLists)[DivTC])!=NULL)) {
    lType = lExp.type; rType = cExp.type;
    do {
      register const TCBlk *pcTCB = ptrTCoper;

      if (lType != NULL) lTypeId = lType->typeId;
      if (rType != NULL) rTypeId = rType->typeId;
      do {
        if (lTypeId==pcTCB->lTypeId || rTypeId==pcTCB->rTypeId) {
          TnameBuf buf;

          bufNameToS(cExp.type->typeId->nameb, buf);
          errWNSS(NoTCAllows | Warn1, lExp.type->typeId->nameb, paramTxt, buf);
          return NULL;}
      } while ((pcTCB = pcTCB->next) != NULL);
      if (lType!=NULL && (lType->rootTyp || (lType = Parent(lType))!=NULL 
					     && !lType->paralTyp)) lType = NULL;
      if (rType!=NULL && (rType->rootTyp || (rType = Parent(rType))!=NULL 
					     && !rType->paralTyp)) rType = NULL;
    } while (lType!=NULL || rType!=NULL);}
  return QuasiNULLval(TpcTypeElt);
}

static void skipToComma(void)
{
  static const Ttok zCommaEndDP[] = {NoSwallowTok, COMMA, ENDDPRAG, ENDDIR,
								       EndSTok};

  skipTok(zCommaEndDP);
  if (curTok.tok == ENDDPRAG) dpragNst++;
}

static void verifyTC(TpcTypeElt lType, TpcTypeElt rType, TpcTypeElt resType)
{
  TtypeSort tsl = intPromotion(lType), tsr = intPromotion(rType), tsRes =
							  intPromotion(resType);
    
#if UINT_MAXC == ULONG_MAXC
  if (tsl<tsr) {TtypeSort tsaux=tsr;; tsr=tsl; tsl=tsaux;}
  if (chkPortbl && tsl==Long && tsr==UInt) errWS(NonPortTC | Warn2, LongStr);
#else
  if (tsl<tsr) tsl = tsr;
#endif 
  if (tsRes!=tsl && !(tsl==Double && tsRes==Float)) {
    Tchar buf[SizTypTxtBuf + 1];

    strcpy(buf, typeToS1(resType));
    errWSTT(IncorTCResTyp, buf, lType, rType);}
}

/* End DCPRAG.C */
